@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program (see the file COPYING); if not, write to the
- * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Free Software Foundation, Inc.,
+ * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
  *
  ****************************************************************
  */
@@ -757,7 +758,7 @@ getttyent()
  *  getlogin() replacement (for SVR4 machines)
  */
 
-# if defined(BUGGYGETLOGIN)
+# if defined(BUGGYGETLOGIN) && defined(UTMP_FILE)
 char *
 getlogin()
 {
